from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Literal, TypeAlias, TypedDict

from ..sparql import Binding, as_long, as_uri, as_string, as_datetime
from ..metacore import UriResource


@dataclass
class DataObjectLite:
	uri: str
	filename: str
	size_bytes: int
	datatype_uri: str
	# station_uri: str | None
	submission_time: datetime
	time_start: datetime
	time_end: datetime

CategorySelector: TypeAlias = str | UriResource | list[str] | list[UriResource] | None
TemporalProp = Literal["submTime", "timeStart", "timeEnd"]
IntegerProp = Literal["size"]
StringProp = Literal["fileName"]
OrderByProp: TypeAlias = TemporalProp | IntegerProp | StringProp
ExclusiveComparator: TypeAlias = Literal["<", ">"]
Comparator: TypeAlias = ExclusiveComparator | Literal["<=", ">=", "="]

class Filter(ABC):
	@abstractmethod
	def render(self) -> str: pass

@dataclass
class TimeFilter(Filter):
	prop: TemporalProp
	op: ExclusiveComparator
	value: datetime | str

	def render(self) -> str:
		v = self.value
		dtStr = v.isoformat() if type(v) is datetime else str(v)
		return f'?{self.prop} {self.op} "{dtStr}"^^xsd:dateTime'

@dataclass
class SizeFilter(Filter):
	op: Comparator
	value: int

	def render(self) -> str:
		return f'?size {self.op} "{self.value}"^^xsd:integer'

class SortProp(TypedDict):
	prop: OrderByProp
class OrderBy(SortProp, total = False):
	descending: bool

def _order_clause(order: OrderBy | OrderByProp | None) -> str:
	if order is None:
		return ""
	if type(order) is str:
		return f"order by ?{order}\n"
	if type(order) is dict:
		prop = order['prop']
		return f"order by desc(?{prop})\n" if order.get("descending") else f"order by ?{prop}\n"
	else:
		return ""

def _selector_values_clause(varname: str, uris: list[str]) -> str:
	if len(uris) == 0:
		return ""
	else:
		return f"VALUES ?{varname} {{ {'<' + '> <'.join(uris) + '>'} }}"

def _get_uri_list(selector: CategorySelector) -> list[str]:
	if selector is None: return []
	def to_uri(uri_or_res: object) -> str:
		return getattr(uri_or_res, "uri", None) or str(uri_or_res)
	if type(selector) is list:
		return [to_uri(uri_or_res) for uri_or_res in selector]
	return [to_uri(selector)]

def dataobj_lite_list(
	datatype: CategorySelector,
	station: CategorySelector,
	filters: list[Filter],
	include_deprecated: bool,
	order_by: OrderBy | OrderByProp | None,
	limit: int,
	offset: int
) -> str:

	specUris =  _get_uri_list(datatype)
	stationUris = _get_uri_list(station)
	stationMandatory = "?dobj cpmeta:wasAcquiredBy/prov:wasAssociatedWith ?station ."
	stationLink = f"OPTIONAL{{ {stationMandatory} }}" if len(stationUris) == 0 else stationMandatory

	filter_clauses = "" if len(filters) == 0 else "FILTER(" + " && ".join([f.render() for f in filters]) + ")"

	return f"""
prefix cpmeta: <http://meta.icos-cp.eu/ontologies/cpmeta/>
prefix prov: <http://www.w3.org/ns/prov#>
prefix xsd: <http://www.w3.org/2001/XMLSchema#>
select ?dobj ?spec ?fileName ?size ?submTime ?timeStart ?timeEnd
where {{
	{_selector_values_clause('spec', specUris)}
	?dobj cpmeta:hasObjectSpec ?spec .
	{_selector_values_clause('station', stationUris)}
	# {stationLink}
	?dobj cpmeta:hasSizeInBytes ?size .
	?dobj cpmeta:hasName ?fileName .
	?dobj cpmeta:wasSubmittedBy/prov:endedAtTime ?submTime .
	?dobj cpmeta:hasStartTime | (cpmeta:wasAcquiredBy / prov:startedAtTime) ?timeStart .
	?dobj cpmeta:hasEndTime | (cpmeta:wasAcquiredBy / prov:endedAtTime) ?timeEnd .
	{"" if include_deprecated else "FILTER NOT EXISTS {[] cpmeta:isNextVersionOf ?dobj}"}
	{filter_clauses}
}}
{_order_clause(order_by)}offset {offset} limit {min(limit, 10000)}"""

def parse_dobj_lite(row: Binding) -> DataObjectLite:
	return DataObjectLite(
		uri = as_uri("dobj", row),
		filename = as_string("fileName", row),
		size_bytes = as_long("size", row),
		datatype_uri = as_uri("spec", row),
		submission_time = as_datetime("submTime", row),
		time_start = as_datetime("timeStart", row),
		time_end = as_datetime("timeEnd", row)
	)



from .envri import EnvriConfig
from .auth import ConfigFileAuth, PasswordAuth, TokenAuth
from .metaclient import MetadataClient
from .dataclient import DataClient
from typing import Tuple

def fromPasswordFile(
	conf: EnvriConfig, conf_file_path: str | None = None
) -> Tuple[ConfigFileAuth, MetadataClient, DataClient]:
	auth = ConfigFileAuth(conf, conf_file_path)
	meta = MetadataClient(conf)
	data = DataClient(conf, auth)
	return auth, meta, data

def fromCookieToken(token: str, conf: EnvriConfig) -> Tuple[MetadataClient, DataClient]:
	auth = TokenAuth(token)
	meta = MetadataClient(conf)
	data = DataClient(conf, auth)
	return meta, data

def fromCredentials(user_id: str, password: str, conf: EnvriConfig) -> Tuple[MetadataClient, DataClient]:
	auth = PasswordAuth(user_id, password, conf)
	meta = MetadataClient(conf)
	data = DataClient(conf, auth)
	return meta, data



from dataclasses import dataclass
from typing import Optional
from .metacore import UriResource, Person, Organization, Station

@dataclass
class Role:
    role: UriResource

@dataclass
class Membership:
    org: UriResource
    role: Role
    startDate: Optional[str]
    endDate: Optional[str]

@dataclass
class PersonWithRoles(Person):
    roles: list[Membership]

@dataclass
class Employee:
    person: Person
    role: Role

@dataclass
class OrgWithStaff(Organization):
    staff: list[Employee]

@dataclass
class StationWithStaff(Station):
    staff: list[Employee]



import requests
from .envri import EnvriConfig
from .sparql import SparqlResults, sparql_select as sparql_select_generic
from .queries.speclist import dobj_spec_lite_list, parse_dobj_spec_lite, DobjSpecLite
from .queries.dataobjlist import DataObjectLite, parse_dobj_lite, dataobj_lite_list
from .queries.dataobjlist import OrderBy, OrderByProp, Filter, TimeFilter, SizeFilter, CategorySelector
from .queries.stationlist import station_lite_list, parse_station, StationLite
from .metacore import DataObject, CPJson, parse_cp_json, DataObjectSpec
from .rolemeta import StationWithStaff
from typing import Type, Literal


class MetadataClient:
	def __init__(self, envri_conf: EnvriConfig):
		self._envri_conf = envri_conf

	def sparql_select(self, query: str, disable_cache: bool = False) -> SparqlResults:
		endpoint = self._envri_conf.sparql_endpoint
		return sparql_select_generic(endpoint, query, disable_cache)

	def list_datatypes(self) -> list[DobjSpecLite]:
		"""
		List available data types
		"""
		query = dobj_spec_lite_list(self._envri_conf)
		qres = self.sparql_select(query)
		return [parse_dobj_spec_lite(b) for b in qres.bindings]

	def list_stations(self, of_station_type_uri: str | None | Literal[False] = None ) -> list[StationLite]:
		"""
		List basic information about available stations. Presense of data from/for these stations is not guaranteed.

		:param of_station_type_uri:
			The URI of the type of interesting station (in RDF metadata, the rdf:type of the station resource, or a supertype of the interesting types).
			If omitted, a default type specific to the ENVRI context is used (e.g. only the ICOS stations for ICOS ENVRI)
			To get all stations without any filtering, use False.
		:return:
			The list of StationLite objects
		"""

		station_type: str | None = None
		if of_station_type_uri is None:
			station_type = self._envri_conf.default_station_type_url
		elif type(of_station_type_uri) == str:
			station_type = of_station_type_uri

		query = station_lite_list(station_type, self._envri_conf)
		qres = self.sparql_select(query)
		return [parse_station(b) for b in qres.bindings]

	def list_data_objects(
		self,
		datatype: CategorySelector = None,
		station: CategorySelector = None,
		filters: list[Filter] = [],
		include_deprecated: bool = False,
		order_by: OrderBy | OrderByProp | None = {"prop": "submTime", "descending": True},
		limit: int = 100,
		offset: int = 0
	) -> list[DataObjectLite]:
		query = dataobj_lite_list(datatype, station, filters, include_deprecated, order_by, limit, offset)
		qres = self.sparql_select(query)
		return [parse_dobj_lite(b) for b in qres.bindings]
	
	def get_datatype_meta(self, dtype: str | DobjSpecLite) -> DataObjectSpec:
		dtype_uri: str
		
		if type(dtype) == str: dtype_uri = dtype
		elif type(dtype) == DobjSpecLite: dtype_uri = dtype.uri
		else: raise ValueError("Dtype must be either landing page URL or an instance of DobjSpecLite")

		return _get_json_meta(dtype_uri, DataObjectSpec)

	def get_dobj_meta(self, dobj: str | DataObjectLite) -> DataObject:
		dobj_uri: str

		if type(dobj) == str: dobj_uri = dobj
		elif type(dobj) == DataObjectLite: dobj_uri = dobj.uri
		else: raise ValueError("Dobj must be either landing page URL or an instance of DataObjectLite")

		return _get_json_meta(dobj_uri, DataObject)
	
	def get_station_meta(self, station: str | StationLite) -> StationWithStaff:
		station_uri: str

		if type(station) == str: station_uri = station
		elif type(station) == StationLite: station_uri = station.uri
		else: raise ValueError("Station must be either landing page URL or an instance of StationLite")

		return _get_json_meta(station_uri, StationWithStaff)

def _get_json_meta(url: str, data_class: Type[CPJson]) -> CPJson:
	resp = requests.get(url = url, headers={"Accept": "application/json"})
	return parse_cp_json(resp.text, data_class=data_class)



from .icos import auth, meta
from .metaclient import TimeFilter, SizeFilter

def init_authentication() -> None:
	return auth.init_config_file()

def list_filtered_atc_co2():
	return meta.list_data_objects(
		datatype = "http://meta.icos-cp.eu/resources/cpmeta/atcCo2Product",
		filters = [TimeFilter("submTime", ">", "2022-05-05T12:00:00Z"), SizeFilter(">", 100000)],
		include_deprecated = False,
		order_by = "size",
		limit = 100
	)

def list_all_stations_in_icos_cp():
	return meta.list_stations(of_station_type_uri=False)



import os
import re
import requests
import io
import shutil
from urllib.parse import urlsplit, unquote
from typing import Iterator
from .queries.dataobjlist import DataObjectLite
from .envri import EnvriConfig
from .auth import AuthTokenProvider
from typing import Tuple, Any


class DataClient:
	def __init__(self, conf: EnvriConfig, auth: AuthTokenProvider) -> None:
		self._conf = conf
		self._auth = auth

	def get_file_stream(self, dobj: str | DataObjectLite) -> Tuple[str, io.BufferedReader]:
		"""
		Fetches the original verbatim content of a data- or document object.

		:param dobj:
			the landing page URI of the object, or a DataObjectLite instance
		:returns:
			a tuple of object filename and a file-like object (io.BufferedReader) with contents,
			which can be either saved to disk or directly sent for processing
		"""

		dobj_uri = to_dobj_uri(dobj)
		url = self._conf.data_service_base_url + urlsplit(dobj_uri).path
		resp = self._auth_get(url, 'fetching data object')

		filename: str
		if type(dobj) is DataObjectLite:
			filename = dobj.filename
		else:
			disp_head = resp.headers["Content-Disposition"]
			fn_match = re.search(r'filename="(.*)"', disp_head)
			if not fn_match:
				raise Exception(f"No filename information in response from {url}")
			filename = unquote(fn_match.group(1))

		return filename, response_as_reader(resp)

	def save_to_folder(self, dobj_uri: str | DataObjectLite, folder_path: str) -> str:
		"""
		Downloads the original verbatim content of a data- or document object,
		and saves it to a folder in a file named according to object's metadata.
		If the file already exists, it gets overwritten.

		:param dobj_uri:
			the landing page URI of the object, or a DataObjectLite instance
		:param folder_path:
			path to the target folder

		:returns:
			the name of the file used to save the object
		"""

		filename, resp_buffer = self.get_file_stream(dobj_uri)

		save_path = os.path.join(folder_path, filename)
		with resp_buffer as reader:
			with open(save_path, "wb") as writer:
				shutil.copyfileobj(reader, writer)
				return filename


	def get_csv_byte_stream(
		self,
		dobj: str | DataObjectLite,
		cols: list[str] = [],
		limit: int | None = None,
		offset: int | None = None
	) -> io.BufferedReader:
		"""
		Fetches a standardized plain CSV serialization of a tabular data object (that has columnar metadata and has been ingested by the data portal).

		:param dobj:
			the landing page URI of the data object, or a DataObjectLite instance
		:param cols:
			list of columns to be included; if omitted, all known columns will be returned
		:param limit:
			limits the number of table rows to be returned
		:param offset:
			number of rows to skip
		"""

		dobj_hash = to_dobj_uri(dobj).split('/')[-1]

		resp = self._auth_get(
			url = self._conf.data_service_base_url + "/csv/" + dobj_hash,
			error_hint = 'fetching CSV',
			params = {
				"col": cols,
				"offset": offset,
				"limit": limit
			}
		)

		return response_as_reader(resp)


	def _auth_get(self, url: str, error_hint: str, params: dict[str, Any] | None = None) -> requests.Response:
		headers = {"Cookie": self._auth.get_token().cookie_value}
		resp = requests.get(url=url, headers=headers, stream=True, params=params)
		if resp.status_code != 200:
			raise Exception(f"Failed {error_hint} from {url}, got response {resp.text}")
		return resp


class HttpResponseStream(io.RawIOBase):
	def __init__(self, resp: requests.Response, chunk_size: int):
		self._http_resp = resp
		self._chunk_size = chunk_size
		self._iterable: Iterator[bytes] = resp.iter_content(chunk_size = chunk_size)
		self.leftover: bytes | None = None
	def readable(self):
		return True
	def readinto(self, b):
		try:
			l = len(b)
			chunk = self.leftover or next(self._iterable)
			output, self.leftover = chunk[:l], chunk[l:]
			b[:len(output)] = output
			return len(output)
		except StopIteration:
			return 0
	def close(self) -> None:
		self._http_resp.close()
		return super().close()

def to_dobj_uri(dobj: str | DataObjectLite) -> str:
	if type(dobj) == DataObjectLite: return dobj.uri
	elif type(dobj) == str: return dobj
	else: raise ValueError('dobj_uri must be a string or a DataObjectLite instance')

def response_as_reader(resp: requests.Response) -> io.BufferedReader:
	chunk_size = io.DEFAULT_BUFFER_SIZE
	return io.BufferedReader(HttpResponseStream(resp, chunk_size), chunk_size)



from datetime import datetime
from typing import TypeAlias, Optional, Any
from dataclasses import dataclass
import re
import requests

@dataclass
class BoundUri:
	uri: str

@dataclass
class BoundLiteral:
	value: str
	datatype: Optional[str]

BoundValue: TypeAlias = BoundUri | BoundLiteral

Binding: TypeAlias = dict[str, BoundValue]

@dataclass
class SparqlResults:
	variable_names: list[str]
	bindings: list[Binding]

def lookup(varname: str, binding: Binding) -> BoundValue:
	v = binding.get(varname)
	if not v:
		raise ValueError(f"Variable {varname} had no bound values in SPARQL response")
	else:
		return v

def lookup_literal_value(varname: str, datatype: Optional[str], binding: Binding) -> str:
	bv = lookup(varname, binding)
	if type(bv) is BoundLiteral:
		if not(datatype) or bv.datatype == datatype:
			return bv.value
		else:
			raise _type_error(varname, f"literal datatype {datatype}", bv)
	else:
		raise _type_error(varname, "a literal", bv)

def as_int(varname: str, binding: Binding) -> int:
	int_str = lookup_literal_value(varname, "http://www.w3.org/2001/XMLSchema#integer", binding)
	return int(int_str)

def as_double(varname: str, binding: Binding) -> float:
	long_str = lookup_literal_value(varname, "http://www.w3.org/2001/XMLSchema#double", binding)
	return float(long_str)

def as_long(varname: str, binding: Binding) -> int:
	long_str = lookup_literal_value(varname, "http://www.w3.org/2001/XMLSchema#long", binding)
	return int(long_str)

def as_string(varname: str, binding: Binding) -> str:
	return lookup_literal_value(varname, None, binding)

def as_datetime(varname: str, binding: Binding) -> datetime:
	dtStr = lookup_literal_value(varname, "http://www.w3.org/2001/XMLSchema#dateTime", binding)
	return datetime.fromisoformat(re.sub(r'Z$', '+00:00', dtStr))


def as_uri(varname: str, binding: Binding) -> str:
	bv = lookup(varname, binding)
	if type(bv) is BoundUri:
		return bv.uri
	else:
		raise _type_error(varname, "a uri value", bv)

def as_opt_double(varname: str, binding: Binding) -> float | None:
	if varname in binding.keys():
		return as_double(varname, binding)
	else: return None

def as_opt_str(varname: str, binding: Binding) -> str | None:
	if varname in binding.keys():
		return as_string(varname, binding)
	else: return None

def as_opt_uri(varname: str, binding: Binding) -> str | None:
	if varname in binding.keys():
		return as_uri(varname, binding)
	else: return None

def _type_error(varname: str, expected: str, bv: BoundValue) -> ValueError:
	msg = f"Was expecting {expected}, got value {bv} for variable {varname} in SPARQL results"
	return ValueError(msg)

def get_sparql_select_json(endpoint: str, query: str, disable_cache: bool) -> Any:
	headers: dict[str, str] = {"Accept": "application/json"}
	if disable_cache:
		headers["Cache-Control"] = "no-cache"
		headers["Pragma"] = "no-cache"
	res = requests.post(url = endpoint, headers=headers, data=bytes(query, "utf-8"))
	if res.status_code != 200:
		raise Exception(f"SPARQL SELECT problem, got response: {res.text}\nThe query was: {query}")
	return res.json()

def sparql_select(endpoint: str, query: str, disable_cache: bool) -> SparqlResults:
	js = get_sparql_select_json(endpoint, query, disable_cache)
	varnames: list[str] = js["head"]["vars"]
	binding_dicts: list[dict[str, Any]] = js["results"]["bindings"]
	bindings = [{key: _parse_bound_value(v) for key, v in b_dict.items()} for b_dict in binding_dicts]
	return SparqlResults(varnames, bindings)

def _parse_bound_value(v: Any) -> BoundValue:
	match v["type"]:
		case "uri": return BoundUri(uri = v["value"])
		case "literal": return BoundLiteral(value = v["value"], datatype=v.get("datatype"))
		case "bnode": raise ValueError("Blank nodes in SPARQL return results are not supported by this" +
			" library. Please re-formulate your query.")
		case bad: raise ValueError(f"Unexpected result type encountered in SPARQL response: {bad}")
